import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, ir, ir_utils, typeinfer, types
from numba.core.ir_utils import compile_to_numba_ir, replace_arg_nodes
from numba.extending import intrinsic
import bodo
import bodo.ir.connector
from bodo import objmode
from bodo.io.fs_io import get_storage_options_pyobject, storage_options_dict_type
from bodo.libs.str_ext import string_type
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.utils.utils import check_and_propagate_cpp_exception, check_java_installation, sanitize_varname


class JsonReader(ir.Stmt):

    def __init__(self, df_out, loc, out_vars, out_types, file_name,
        df_colnames, orient, convert_dates, precise_float, lines,
        compression, storage_options):
        self.connector_typ = 'json'
        self.df_out = df_out
        self.loc = loc
        self.out_vars = out_vars
        self.out_types = out_types
        self.file_name = file_name
        self.df_colnames = df_colnames
        self.orient = orient
        self.convert_dates = convert_dates
        self.precise_float = precise_float
        self.lines = lines
        self.compression = compression
        self.storage_options = storage_options

    def __repr__(self):
        return ('{} = ReadJson(file={}, col_names={}, types={}, vars={})'.
            format(self.df_out, self.file_name, self.df_colnames, self.
            out_types, self.out_vars))


import llvmlite.binding as ll
from bodo.io import json_cpp
ll.add_symbol('json_file_chunk_reader', json_cpp.json_file_chunk_reader)


@intrinsic
def json_file_chunk_reader(typingctx, fname_t, lines_t, is_parallel_t,
    nrows_t, compression_t, bucket_region_t, storage_options_t):
    assert storage_options_t == storage_options_dict_type, "Storage options don't match expected type"

    def codegen(context, builder, sig, args):
        rwqq__dare = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(1), lir.IntType(1), lir.
            IntType(64), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer()])
        xgk__ngnji = cgutils.get_or_insert_function(builder.module,
            rwqq__dare, name='json_file_chunk_reader')
        iayj__utmx = builder.call(xgk__ngnji, args)
        context.compile_internal(builder, lambda :
            check_and_propagate_cpp_exception(), types.none(), [])
        wwcoa__gsjfl = cgutils.create_struct_proxy(types.stream_reader_type)(
            context, builder)
        pikdq__zdhlm = context.get_python_api(builder)
        wwcoa__gsjfl.meminfo = pikdq__zdhlm.nrt_meminfo_new_from_pyobject(
            context.get_constant_null(types.voidptr), iayj__utmx)
        wwcoa__gsjfl.pyobj = iayj__utmx
        pikdq__zdhlm.decref(iayj__utmx)
        return wwcoa__gsjfl._getvalue()
    return types.stream_reader_type(types.voidptr, types.bool_, types.bool_,
        types.int64, types.voidptr, types.voidptr, storage_options_dict_type
        ), codegen


def remove_dead_json(json_node, lives_no_aliases, lives, arg_aliases,
    alias_map, func_ir, typemap):
    uzae__vow = []
    fjkn__siq = []
    sinwr__ere = []
    for upfl__obncz, buc__nymk in enumerate(json_node.out_vars):
        if buc__nymk.name in lives:
            uzae__vow.append(json_node.df_colnames[upfl__obncz])
            fjkn__siq.append(json_node.out_vars[upfl__obncz])
            sinwr__ere.append(json_node.out_types[upfl__obncz])
    json_node.df_colnames = uzae__vow
    json_node.out_vars = fjkn__siq
    json_node.out_types = sinwr__ere
    if len(json_node.out_vars) == 0:
        return None
    return json_node


def json_distributed_run(json_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    if bodo.user_logging.get_verbose_level() >= 1:
        spik__clq = (
            'Finish column pruning on read_json node:\n%s\nColumns loaded %s\n'
            )
        dgj__cef = json_node.loc.strformat()
        lhw__zhhwd = json_node.df_colnames
        bodo.user_logging.log_message('Column Pruning', spik__clq, dgj__cef,
            lhw__zhhwd)
        dmrij__gnby = [nfe__kfqzq for upfl__obncz, nfe__kfqzq in enumerate(
            json_node.df_colnames) if isinstance(json_node.out_types[
            upfl__obncz], bodo.libs.dict_arr_ext.DictionaryArrayType)]
        if dmrij__gnby:
            lorik__exynk = """Finished optimized encoding on read_json node:
%s
Columns %s using dictionary encoding to reduce memory usage.
"""
            bodo.user_logging.log_message('Dictionary Encoding',
                lorik__exynk, dgj__cef, dmrij__gnby)
    parallel = False
    if array_dists is not None:
        parallel = True
        for zzy__hqix in json_node.out_vars:
            if array_dists[zzy__hqix.name
                ] != distributed_pass.Distribution.OneD and array_dists[
                zzy__hqix.name] != distributed_pass.Distribution.OneD_Var:
                parallel = False
    vampp__ofspy = len(json_node.out_vars)
    qjp__nereb = ', '.join('arr' + str(upfl__obncz) for upfl__obncz in
        range(vampp__ofspy))
    vsz__fwi = 'def json_impl(fname):\n'
    vsz__fwi += '    ({},) = _json_reader_py(fname)\n'.format(qjp__nereb)
    kvml__doto = {}
    exec(vsz__fwi, {}, kvml__doto)
    jmfyk__nuz = kvml__doto['json_impl']
    xycy__ipzyq = _gen_json_reader_py(json_node.df_colnames, json_node.
        out_types, typingctx, targetctx, parallel, json_node.orient,
        json_node.convert_dates, json_node.precise_float, json_node.lines,
        json_node.compression, json_node.storage_options)
    gmjpr__rqbf = compile_to_numba_ir(jmfyk__nuz, {'_json_reader_py':
        xycy__ipzyq}, typingctx=typingctx, targetctx=targetctx, arg_typs=(
        string_type,), typemap=typemap, calltypes=calltypes).blocks.popitem()[1
        ]
    replace_arg_nodes(gmjpr__rqbf, [json_node.file_name])
    uyepd__pkbue = gmjpr__rqbf.body[:-3]
    for upfl__obncz in range(len(json_node.out_vars)):
        uyepd__pkbue[-len(json_node.out_vars) + upfl__obncz
            ].target = json_node.out_vars[upfl__obncz]
    return uyepd__pkbue


numba.parfors.array_analysis.array_analysis_extensions[JsonReader
    ] = bodo.ir.connector.connector_array_analysis
distributed_analysis.distributed_analysis_extensions[JsonReader
    ] = bodo.ir.connector.connector_distributed_analysis
typeinfer.typeinfer_extensions[JsonReader
    ] = bodo.ir.connector.connector_typeinfer
ir_utils.visit_vars_extensions[JsonReader
    ] = bodo.ir.connector.visit_vars_connector
ir_utils.remove_dead_extensions[JsonReader] = remove_dead_json
numba.core.analysis.ir_extension_usedefs[JsonReader
    ] = bodo.ir.connector.connector_usedefs
ir_utils.copy_propagate_extensions[JsonReader
    ] = bodo.ir.connector.get_copies_connector
ir_utils.apply_copy_propagate_extensions[JsonReader
    ] = bodo.ir.connector.apply_copies_connector
ir_utils.build_defs_extensions[JsonReader
    ] = bodo.ir.connector.build_connector_definitions
distributed_pass.distributed_run_extensions[JsonReader] = json_distributed_run
compiled_funcs = []


def _gen_json_reader_py(col_names, col_typs, typingctx, targetctx, parallel,
    orient, convert_dates, precise_float, lines, compression, storage_options):
    zxkbw__fph = [sanitize_varname(nfe__kfqzq) for nfe__kfqzq in col_names]
    pplg__gpez = ', '.join(str(upfl__obncz) for upfl__obncz, kbbjp__vgs in
        enumerate(col_typs) if kbbjp__vgs.dtype == types.NPDatetime('ns'))
    uzfe__sdoi = ', '.join(["{}='{}'".format(wvxmi__hqtf, bodo.ir.csv_ext.
        _get_dtype_str(kbbjp__vgs)) for wvxmi__hqtf, kbbjp__vgs in zip(
        zxkbw__fph, col_typs)])
    urwr__sxr = ', '.join(["'{}':{}".format(xhdm__xhph, bodo.ir.csv_ext.
        _get_pd_dtype_str(kbbjp__vgs)) for xhdm__xhph, kbbjp__vgs in zip(
        col_names, col_typs)])
    if compression is None:
        compression = 'uncompressed'
    vsz__fwi = 'def json_reader_py(fname):\n'
    vsz__fwi += '  df_typeref_2 = df_typeref\n'
    vsz__fwi += '  check_java_installation(fname)\n'
    vsz__fwi += f"""  bucket_region = bodo.io.fs_io.get_s3_bucket_region_njit(fname, parallel={parallel})
"""
    if storage_options is None:
        storage_options = {}
    storage_options['bodo_dummy'] = 'dummy'
    vsz__fwi += (
        f'  storage_options_py = get_storage_options_pyobject({str(storage_options)})\n'
        )
    vsz__fwi += (
        '  f_reader = bodo.ir.json_ext.json_file_chunk_reader(bodo.libs.str_ext.unicode_to_utf8(fname), '
        )
    vsz__fwi += (
        """    {}, {}, -1, bodo.libs.str_ext.unicode_to_utf8('{}'), bodo.libs.str_ext.unicode_to_utf8(bucket_region), storage_options_py )
"""
        .format(lines, parallel, compression))
    vsz__fwi += '  if bodo.utils.utils.is_null_pointer(f_reader._pyobj):\n'
    vsz__fwi += "      raise FileNotFoundError('File does not exist')\n"
    vsz__fwi += f'  with objmode({uzfe__sdoi}):\n'
    vsz__fwi += f"    df = pd.read_json(f_reader, orient='{orient}',\n"
    vsz__fwi += f'       convert_dates = {convert_dates}, \n'
    vsz__fwi += f'       precise_float={precise_float}, \n'
    vsz__fwi += f'       lines={lines}, \n'
    vsz__fwi += '       dtype={{{}}},\n'.format(urwr__sxr)
    vsz__fwi += '       )\n'
    vsz__fwi += (
        '    bodo.ir.connector.cast_float_to_nullable(df, df_typeref_2)\n')
    for wvxmi__hqtf, xhdm__xhph in zip(zxkbw__fph, col_names):
        vsz__fwi += '    if len(df) > 0:\n'
        vsz__fwi += "        {} = df['{}'].values\n".format(wvxmi__hqtf,
            xhdm__xhph)
        vsz__fwi += '    else:\n'
        vsz__fwi += '        {} = np.array([])\n'.format(wvxmi__hqtf)
    vsz__fwi += '  return ({},)\n'.format(', '.join(bkvvn__bhyl for
        bkvvn__bhyl in zxkbw__fph))
    imkms__jjch = globals()
    imkms__jjch.update({'bodo': bodo, 'pd': pd, 'np': np, 'objmode':
        objmode, 'check_java_installation': check_java_installation,
        'df_typeref': bodo.DataFrameType(tuple(col_typs), bodo.
        RangeIndexType(None), tuple(col_names)),
        'get_storage_options_pyobject': get_storage_options_pyobject})
    kvml__doto = {}
    exec(vsz__fwi, imkms__jjch, kvml__doto)
    xycy__ipzyq = kvml__doto['json_reader_py']
    zpcv__jhzjk = numba.njit(xycy__ipzyq)
    compiled_funcs.append(zpcv__jhzjk)
    return zpcv__jhzjk
